/**************************************************************************************
*
*	Copyright (C) 2010 - 2012 Brent Andrew O. Chua - See LICENSE.txt for more info.
*		
*	Filename	-	RenderPathForwardPlus
*
*	Description -	A Forward+ render path implementation.
*					This requires a 'x' pass stage.
*					1. Z-prepass
*					2. Light accumulation
*					3. Opaque forward rendering
*					4. ...
*
*	Comments	-	TODO: Make it so that this implementation is derived from 
*						IRenderPass interface.
*
*	Modification History:
*			Name			   Date					Description
*			MrCodeSushi	-	07/23/2013		-	File creation
**************************************************************************************/

#ifndef _PSX_RENDER_PASS_FORWARD_PLUS_H_
#define _PSX_RENDER_PASS_FORWARD_PLUS_H_

#include "PulseSTD.h"
#include "ParameterContainer.h"
#include "RenderEntryList.h"
#include "Array.h"

namespace Pulse
{
	// A simple light representation to be used bu the shader
	// This should be the same as the oneused in core.shx
	struct ShaderLight
	{
		float position[3];
		float color[3];
		float range;
	};

	typedef PSX_Pair< ComponentCamera*, CameraRenderPathSettingsFP* > CameraRPSettings;

	PSX_EnumBegin( EForwadPlusNumLights )
		_128,
		_256,	// Only one currently implemented
		_512,
	PSX_EnumEnd()

	class RenderPathForwardPlus
	{
	public:

		RenderPathForwardPlus( void );

		virtual ~RenderPathForwardPlus( void );

		const UINT GetLightingTileSize( void ) const { return m_lightTileSize; }

		const UINT GetNumLightsPerTile( void ) const { return m_numLights; }

		// Used to get the location of techniques
		const CHAR * GetTechniquesPath( void ) const { return PSX_String("ForwardPlus\\"); }
		
		const CHAR * GetTechniquesLibraryPath( void ) const { return PSX_String("ForwardPlus\\Library.rtl"); }

		// Standard construction of render entries for models
		EErrorCode::Type BuildRenderEntries( const Model *pModel, ComponentModelRenderer *pModelRenderer );
		
		// Build render entries with a render entry list instead of component
		EErrorCode::Type BuildRenderEntries( const Model *pModel, RenderEntryList *pRenderEntries );

		// Render entry construction for lights
		EErrorCode::Type BuildRenderEntries( ComponentLight *pLight );

		// The render path can handle any extra setup before executing the render view
		//	like setting up lights fore deferred rendering.
		void ExecuteRenderView( RenderView *pView );

		//
		EErrorCode::Type CreateCameraSettings( ComponentCamera *pCamera, CameraRenderPathSettingsFP **ppOut );

		EErrorCode::Type RemoveCameraSettings( const ComponentCamera *pCamera, const CameraRenderPathSettingsFP *pSettings );

	protected:

		friend Renderer;

		typedef Array<CameraRPSettings> CameraRPSettingsList;

		virtual EErrorCode::Type Initialize( void );

		virtual void Cleanup( void );

		CameraRPSettings * FindCameraSettings( const ComponentCamera *pCamera );

	private:

		UINT m_lightTileSize;	// Number of light tile size
		UINT m_numLights;

		// Cached lighting instances
		///////////////////////////////

		// Point lights
		//	If inside - draw quad
		//  If outside and intersecting near and far plane - draw quad
		//	If outside and intersecting near only - draw back
		//	If outside and intersecting far only - draw front
		//	(else) If outside and none intersecting - draw front

		// Materials
		Material *m_pPointLightMat;		// point light
		Material *m_pPointLightFront;	// point light
		Material *m_pPointLightBack;	// point light
		Material *m_pPointLightQuad;	// point light
		
		// Geometry
		Geometry *m_pPointLightGeo;
		Geometry *m_pGeoQuad;

		CameraRPSettingsList m_viewRPSettings;
	};
}

#endif /* _PSX_RENDER_PASS_FORWARD_PLUS_H_ */