/**************************************************************************************
*
*	Copyright (C) 2010 - 2012 Brent Andrew O. Chua - See LICENSE.txt for more info.
*		
*	Filename	- IResourcePoolItem.h
*
*	Description - The base resource item class used by the resource pool.
*
*	Comments	- Also contains ResourceCode class that determins the type of resrouce 
*					item.
*
*	Modification History:
*			Name			   Date					Description
*			MrCodeSushi	-	04/10/2010	-	Creation of this file.
**************************************************************************************/

#ifndef _PSX_RESOURCE_POOL_ITEM_H_
#define _PSX_RESOURCE_POOL_ITEM_H_

#include "PulseSTD.h"
#include "IRefCounter.h"
#include "StoragePool.h"
#include "String.h"
#include "BitFlags.h"
#include "AsyncLoadResource.h"

namespace Pulse
{

#define PSX_MakeStoragePoolFriend(className) 	friend class ResourcePool<className>;	\
												friend class Storage<className>;

	class IResourcePool;

	PSX_EnumBegin( EResource )
		UNKNOWN,
		VIDEO,
		AUDIO,
		WORLD,
		CUSTOM,
	PSX_EnumEnd()
	
	PSX_EnumBegin( EVideoResource )
		UNKNOWN,
		VIEWPORT,
		INPUT_LAYOUT,
		BLEND_STATE,
		DEPTH_STENCIL_STATE,
		RASTERIZER_STATE,
		SAMPLER_STATE,
		BUFFER,
		TEXTURE1D,
		TEXTURE2D,
		TEXTURE3D,
		SWAP_CHAIN,
		VERTEX_DECLARATION,
		VERTEX_BUFFER,
		INDEX_BUFFER,
		CONSTANT_BUFFER,
		STRUCTURED_BUFFER,
		SHADER,
		TECHNIQUE,
		EFFECT,
		SHADER_TECHNIQUE,
		MATERIAL,
		GEOMETRY,
		FONT,	
	PSX_EnumEnd()

	PSX_EnumBegin( EWorldResource )
		MODEL,
		SECTOR,
	PSX_EnumEnd()

	class ResourceCode
	{
	public:

		enum {
			INVALID_CODE = 0xFFFFFFFF,
		};

		union
		{
			struct
			{
				WORD type;
				WORD subType;
			};
			DWORD value;
		};

		ResourceCode( void ) : value(INVALID_CODE) { }
		ResourceCode( DWORD _type ) : value( _type ) { }
		ResourceCode( WORD _type, WORD _subType ) : type(_type), subType(_subType) { }

		void SetInvalid( void ) { value = INVALID_CODE; }

		const BOOL IsValid( void ) { return value != INVALID_CODE; }

		const BOOL operator < ( const ResourceCode &rhs ) const { return value < rhs.value; }
		const BOOL operator > ( const ResourceCode &rhs ) const { return value > rhs.value; }
		const BOOL operator == ( const ResourceCode &rhs ) const { return value == rhs.value; }
		const BOOL operator != ( const ResourceCode &rhs ) const { return !(value == rhs.value); }

	};

	struct BaseResDesc
	{
		BaseResDesc( const CHAR *pName = PSX_String("Unnamed") ) 
			: m_name(pName),bAsyncLoad( FALSE )
		{
		}

		virtual ~BaseResDesc( void ) { }

		void SetAsyncLoad( BOOL _bAsyncLoad ) { bAsyncLoad = _bAsyncLoad; } 

		BOOL IsAsyncLoad( void ) { return bAsyncLoad; }

		virtual const ResourceCode GetType( void ) const = 0;

		String m_name;
		BOOL bAsyncLoad;

	};

	// NOTE: Some Guidelines on the Create, Destroy, Disable, Restore Resources.
	//	Create and Destroy are now the constructor and destructors. 
	//	Minimize code in your constructor and place them in CreateResource.
	//	Minimize code in your destructor and place them in DestroyResource.
	//	Graphics resources sometimes gets broken(DX9). Call UnloadResources to release 
	//	important resources; After calling UnloadResource call LoadResource to 
	//	reload the resource again.

	class IResourcePoolItem : public RefCounter
	{
	public:

		enum RESOURCE_FLAG
		{
			RESOURCE_FLAG_CREATED = 0,
			RESOURCE_FLAG_LOADED,
			RESOURCE_FLAG_DISABLED,
			RESOURCE_FLAG_ALTERED,
			RESOURCE_FLAG_ERROR,
		};

		IResourcePoolItem( void );			// Called on creation.
		virtual ~IResourcePoolItem( void );	// Called on destruction

		// These defines the base interface for all derived resources
		virtual BOOL CreateResource( BaseResDesc *pDesc )	{ SetFlagCreated();		return TRUE;	}	// Your new Constructor
		virtual BOOL DestroyResource( void )				{ SetFlagDestroyed();	return TRUE;	}	// Your new destructor
		virtual BOOL LoadResource( void )					{ SetFlagLoaded();		return TRUE;	}	// Reloads/Initialize your resourcess
		virtual BOOL UnloadResource( void )					{ SetFlagUnloaded();	return TRUE;	}	// Disables/Releases resources
		virtual BOOL RestoreResource( void )				{ SetFlagRestored();	return TRUE;	}
		
		// NOTE: Newly added. Used to save resource
		virtual BOOL Saveresource( void )					{ SetFlagSaved();		return TRUE;	}

		virtual const ResourceCode	GetType( void ) const = 0;
		ResourceCode			GetResourceType( void ) const { return GetType(); }
		IResourcePool *			GetResourcePool( void ) const;
		HRes					GetResourceHandle( void ) const;
		BitFlagsU32				GetResourceFlags( void ) const;

		BOOL IsCreated( void ) const;
		BOOL IsDisabled( void ) const;
		BOOL IsLoaded( void ) const;
		BOOL IsError( void ) const; // If asyncload and load has failed.
		const String * GetResourceName( void ) const;
		
		void SetResourceName( const CHAR *pName );
		void SetFlagAltered( bool bAltered );
		BOOL GetFlagAltered( void );

		virtual INT Release( void );

	protected:

		// NOTE: Derived classes calls this method to perform initialization.
		// It performs syncronous or asynchronous loading based on desc's bAsyncLoad.
		EErrorCode::Type PerformInitialize( BOOL bAsyncLoad );

	protected:

		void SetResourcePool( IResourcePool *pResPool );
		void SetResourceHandle( HRes hRes );
		void SetResourceFlag( DWORD flagBit, BOOL bOn );

		void SetFlagCreated( void );
		void SetFlagDisabled( void );
		void SetFlagLoaded( void );
		void SetFlagUnloaded( void );
		void SetFlagRestored( void );
		void SetFlagDestroyed( void );
		void SetFlagSaved( void );
		void SetFlagError( void );
		

	private:

		template < typename T >
		friend class ResourcePool;

		IResourcePoolItem( const IResourcePoolItem & );
		IResourcePoolItem & operator = ( const IResourcePoolItem & rhs );

		IResourcePool			*m_pResPool;
		HRes					m_hRes;
		BitFlagsU32				m_resFlags;

	protected:

		// Adapter for the async resourece loading functionality
		class AdapterResPoolLoader : public AdapterResourceLoader
		{
		public:

			AdapterResPoolLoader( void );

			virtual ~AdapterResPoolLoader( void );

			void SetResourcePoolItem( IResourcePoolItem *pRes );

			virtual BOOL Run( TaskStatus *pTaskStatus );

		private:

			IResourcePoolItem *m_pRes;
		};

		AdapterResPoolLoader m_adapterResLoader;
		AsyncLoadResource m_asyncLoad;
	};

	PSX_INLINE IResourcePoolItem::~IResourcePoolItem( void )
	{
		PSX_Assert( !IsCreated(), "Failed to destroy resource." );
	}

	PSX_INLINE IResourcePool *	IResourcePoolItem::GetResourcePool( void ) const
	{
		return m_pResPool;
	}

	PSX_INLINE HRes IResourcePoolItem::GetResourceHandle( void ) const
	{
		return m_hRes;
	}

	PSX_INLINE BitFlagsU32 IResourcePoolItem::GetResourceFlags( void ) const
	{
		return m_resFlags;
	}

	PSX_INLINE BOOL IResourcePoolItem::IsCreated( void ) const
	{
		return m_resFlags.TestBit( RESOURCE_FLAG_CREATED );
	}

	PSX_INLINE BOOL IResourcePoolItem::IsDisabled( void ) const
	{
		return m_resFlags.TestBit( RESOURCE_FLAG_DISABLED );
	}

	PSX_INLINE BOOL IResourcePoolItem::IsLoaded( void ) const
	{
		return m_resFlags.TestBit( RESOURCE_FLAG_LOADED );
	}

	PSX_INLINE BOOL IResourcePoolItem::IsError( void ) const
	{
		return m_resFlags.TestBit( RESOURCE_FLAG_ERROR );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagAltered( bool bAltered )
	{
		m_resFlags.SetBit( RESOURCE_FLAG_ALTERED, bAltered );
	}

	PSX_INLINE BOOL IResourcePoolItem::GetFlagAltered( void )
	{
		return m_resFlags.TestBit( RESOURCE_FLAG_ALTERED );
	}

	PSX_INLINE void IResourcePoolItem::SetResourcePool( IResourcePool *pResPool )
	{
		m_pResPool = pResPool;
	}

	PSX_INLINE void IResourcePoolItem::SetResourceHandle( HRes hRes )
	{
		m_hRes = hRes;
	}

	PSX_INLINE void IResourcePoolItem::SetResourceFlag( DWORD flagBit, BOOL bOn )
	{
		m_resFlags.SetBit( flagBit, bOn );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagCreated( void )
	{
		SetResourceFlag( RESOURCE_FLAG_CREATED, TRUE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagDisabled( void )
	{
		SetResourceFlag( RESOURCE_FLAG_DISABLED, TRUE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagLoaded( void )
	{
		SetFlagAltered( FALSE );
		SetResourceFlag( RESOURCE_FLAG_LOADED, TRUE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagUnloaded( void )
	{
		SetResourceFlag( RESOURCE_FLAG_LOADED, FALSE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagRestored( void )
	{
		SetResourceFlag( RESOURCE_FLAG_DISABLED, FALSE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagDestroyed( void )
	{
		SetResourceFlag( RESOURCE_FLAG_CREATED, FALSE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagSaved( void )
	{
		SetFlagAltered( FALSE );
	}

	PSX_INLINE void IResourcePoolItem::SetFlagError( void )
	{
		SetResourceFlag( RESOURCE_FLAG_ERROR, TRUE );
	}

}

#endif /* _PSX_RESOURCE_POOL_ITEM_H_ */