/**************************************************************************************
*
*	Copyright (C) 2009 - 2012 Brent Andrew O. Chua - See LICENSE.txt for more info.
*		
*	Filename	-	SmartPointer.h
*
*	Description -	SmartPointer is a template class that provides automatic 
*						garbage collection for the heap allocated data.
*
*	Comments	-	NEVER EVER pass a dynamically allocated object that's managed by 
*						a SmartPointer to another SmartPointer.
*					Ex. TestClass *pObj = new TestClass;
*						SmartPointer< TestClass > smart1(pObj);
*						SmartPointer< TestClass > smart2(pObj); // Will crash!
*
*				-	Never let two SmartPointers point to each other.
*					TODO: Evaluate the need of SmartPointerArr. And if the methods
*						really needs to be virtuals...
*
*	Modification History:
*			Name			   Date					Description
*			MrCodeSushi	-	10/10/2009	-	File creation and base implementation.
*			MrCodeSushi	-	10/11/2009	-	Array support ( SmartPointerArr<T> ).
*			MrCodeSushi	-	05/22/2010	-	Now using AtomicInt for m_refCount(Multi-Threading).
*			MrCodeSushi	-	08/26/2013	-	Adding WeakPointer
**************************************************************************************/

#ifndef _PSX_SMARTPOINTER_H_
#define _PSX_SMARTPOINTER_H_

#include "PulseSTD.h"
//#include "IRefCounter.h"
#include "AtomicInt.h"

namespace Pulse
{

	template < typename T >
	class SmartPointer
	{
	public:

		typedef T * PointerType;

		SmartPointer( void );
		SmartPointer( const SmartPointer< T > &smartPointer );
		SmartPointer( const T *pData );
		~SmartPointer();

		const T & operator * ( void ) const;
		T & operator * ( void );
		T * operator -> ( void ) const;

		operator T * ( void ) const;
		operator T & ( void ) const;

		BOOL operator == ( const SmartPointer< T > &rhs ) const;

		SmartPointer & operator = ( const SmartPointer< T > &rhs );
		SmartPointer & operator = ( const T *pData );

		void Reset( const T *pData );
		void Reset( void );

		const SIZE_T GetRefCount( void ) const;

		const BOOL IsNull( void ) const;

		const BOOL IsValid( void ) const { return IsNull() == FALSE; }

	protected:

		void Release();
		void AddRef();

		//class ISmartData : public IRefCounter
		//{
		//public:
		//	virtual void AddRef( void ) = 0;
		//	virtual void Release( void ) = 0;

		//	virtual T * GetData( void ) = 0;
		//	virtual const T * GetData() const = 0;
		//	virtual void SetData( T *pData ) = 0;

		//	//virtual ~ISmartData( void ) { }; // Eveything should be handled by Release()...
		//};

		friend WeakPointer<T>;

		class SmartData /*: public ISmartData*/
		{
		public:

			SmartData( const T* pData );

			/*virtual*/ void AddRef( void );
			/*virtual*/ void Release( void );

			void AddWeakRef( void );
			void ReleaseWeakRef( void );

			/*virtual*/ T * GetData( void );
			/*virtual*/ const T * GetData( void ) const;
			///*virtual*/ void SetData( T *pData ); // BAD idea

		private:

			void DeleteData( void );

		private:

			PSX_NO_CALL SmartData();

		protected:

			friend SmartPointer;
			friend WeakPointer<T>;

			T			*m_pData;
			AtomicInt	m_refCount;
			AtomicInt	m_weakCount;	// Default is one as long as there are smartpointers.
		};

		explicit SmartPointer( SmartData &smartData );


		SmartData *m_pSmartObj;
	};

	template < typename T >
	class WeakPointer
	{
	public:

		typedef T * PointerType;

		WeakPointer( void );
		WeakPointer( const SmartPointer< T > &smartPointer );
		~WeakPointer();

		SmartPointer<T> Lock( void );

		BOOL operator == ( const WeakPointer< T > &rhs ) const;

		WeakPointer & operator = ( const WeakPointer< T > &rhs );
		WeakPointer & operator = ( const SmartPointer< T > &smartPointer );

		void Reset( const SmartPointer< T > &smartPointer );
		void Reset( void );

		const SIZE_T GetRefCount( void ) const;
		const SIZE_T GetWeakCount( void ) const;

		const BOOL IsNull( void ) const;

		const BOOL IsValid( void ) const { return IsNull() == FALSE; }

	protected:

		void Release();
		void AddRef();

		typedef typename SmartPointer<T>::SmartData SmartData;

		SmartData *m_pSmartObj;
	};

	// SmartPointer for handling arrays
	// NOTE: It doesn't look like we're using this. So in case we do 
	//	 end up using this, make sure to check this class first.
	template < typename T >
	class SmartPointerArr : public SmartPointer< T >
	{
	public:

		SmartPointerArr( void );
		SmartPointerArr( SmartPointerArr< T > &smartPointerArr );
		explicit SmartPointerArr( T *pData );
		~SmartPointerArr();

		SmartPointerArr & operator = ( const T *pData );
		SmartPointerArr & operator = ( const SmartPointerArr< T > &rhs );

	private:

		class SmartDataArr : public SmartData
		{
		public:
			SmartDataArr( void );
			SmartDataArr( const T* pData );

			virtual void Release( void );
		};
	};
}

#include "SmartPointer.inl"

#endif _PSX_SMARTPOINTER_H_
