/**************************************************************************************
*
*	Copyright (C) 2009 - 2012 Brent Andrew O. Chua - See LICENSE.txt for more info.
*		
*	Filename	-	Buffer.cpp
*
*	Comments	-	See Buffer.h
*
**************************************************************************************/
#include "../Include/Buffer.h"
#include "../Include/Engine.h"
#include "../Include/AsyncLoadResource.h"
#include "../Include/GraphicsMapping.h"

namespace Pulse
{
	BufferDesc::BufferDesc( void )
	{
		SetDefault();
	}

	void BufferDesc::SetDefault( void )
	{
		m_format		= EGraphicsFormat::UNKNOWN;
		m_numElements	= 0;
		m_usage			= EBufferUsage::UNKNOWN;
		m_bAllowRawView = FALSE;
		m_pData			= PSX_NULL;

		m_srvDesc.SetDefaults();
		m_uavDesc.SetDefaults();
	}

	void BufferDesc::Set( EGraphicsFormat::Type	format, SIZE_T numElements, EBufferUsage::Type usage, void *pInitialData, BOOL bAllowRawView )
	{
		m_format = format;
		m_numElements = numElements;
		m_usage = usage;
		m_bAllowRawView = bAllowRawView;
		m_pData = pInitialData;

		{
			BufferInfoSRV info;

			info.firstElement = 0;
			info.numElements = numElements;

			m_srvDesc.SetDefaults();
			m_srvDesc.Set( format, &info );
		}

		{
			BufferInfoUAV info;

			info.firstElement = 0;
			info.numElements = numElements;
			info.flags = bAllowRawView ? EBufferUAVFlags::RAW : 0;

			m_uavDesc.SetDefaults();
			m_uavDesc.Set( format, &info );
		}
	}

	Buffer::Buffer( void )
	{
	}

	Buffer::~Buffer( void )
	{

	}

	BOOL Buffer::CreateResource( BaseResDesc *pDesc )	
	{
		if ( pDesc->GetType() != GetType() )
			return FALSE;

		if ( ((BufferDesc*)pDesc)->m_usage == EBufferUsage::STATIC &&
			((BufferDesc*)pDesc)->m_pData == PSX_NULL )
		{
			PSX_PushError( "Can't create static buffer without providing an initial data." );
			return FALSE;
		}

		m_desc = *((BufferDesc*)pDesc);

		if ( PerformInitialize( m_desc.IsAsyncLoad() ) == EErrorCode::OKAY )
			SetFlagCreated();

		return IsCreated();	
	}

	BOOL Buffer::DestroyResource( void )				
	{
		UnloadResource();
		SetFlagDestroyed();	
		return TRUE;	
	}

	BOOL Buffer::LoadResource( void )				
	{ 
		if ( Initialize() == EErrorCode::OKAY )
		{
			SetFlagLoaded();	
			return TRUE;	
		}

		return FALSE;
	}

	BOOL Buffer::UnloadResource( void )				
	{
		Cleanup();
		SetFlagUnloaded();
		return FALSE;	
	}

	EErrorCode::Type Buffer::Initialize( void )
	{
		Renderer *pRenderer = Engine::GetInstance()->GetRenderer();
		ID3D11Device *pDevice = (ID3D11Device*)pRenderer->GetNakedDevice();

		HRESULT hr;
		D3D11_BUFFER_DESC desc;

		desc.ByteWidth				= (UINT)GraphicsMapping::ToGraphicsFormatSize(m_desc.m_format) * m_desc.m_numElements;
		desc.BindFlags				= D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS;
		desc.MiscFlags				= m_desc.m_bAllowRawView ? D3D11_RESOURCE_MISC_BUFFER_ALLOW_RAW_VIEWS : 0;
		desc.StructureByteStride	= 0;

		if ( m_desc.m_usage == EBufferUsage::DYNAMIC )
		{
			desc.Usage = D3D11_USAGE_DYNAMIC;
			desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
		}
		else if ( m_desc.m_usage == EBufferUsage::STATIC )
		{
			desc.Usage = D3D11_USAGE_IMMUTABLE;
			desc.CPUAccessFlags = 0;
		}
		else if ( m_desc.m_usage == EBufferUsage::DEFAULT )
		{
			desc.Usage = D3D11_USAGE_DEFAULT;
			desc.CPUAccessFlags = 0;
		}
		else if ( m_desc.m_usage == EBufferUsage::STAGING )
		{
			desc.Usage = D3D11_USAGE_STAGING;
			desc.CPUAccessFlags = D3D11_CPU_ACCESS_READ | D3D11_CPU_ACCESS_WRITE;
		}
		else
		{
			PSX_PushError( "Not yet implemented." );
		}

		if ( m_desc.m_pData )
		{
			D3D11_SUBRESOURCE_DATA data;
			data.pSysMem = m_desc.m_pData;
			data.SysMemPitch = 0;
			data.SysMemSlicePitch = 0;

			hr = pDevice->CreateBuffer( &desc, &data, &m_pBuffer );
		}
		else
		{
			hr = pDevice->CreateBuffer( &desc, PSX_NULL, &m_pBuffer );
		}

		m_desc.m_pData = PSX_NULL;

		if ( FAILED(hr) )
		{
			PSX_PushError( "Failed to create D3D11 structured buffer." );
			return EErrorCode::GRAPHICS;
		}

		CreateResourceViews( m_pBuffer, desc.BindFlags, PSX_NULL, &m_desc.m_srvDesc, &m_desc.m_uavDesc );

		return EErrorCode::OKAY;
	}

	void Buffer::Cleanup( void )
	{
		ReleaseResourceViews();
		ReleaseBuffer();
	}
}