/**************************************************************************************
*
*	Copyright (C) 2009 - 2012 Brent Andrew O. Chua - See LICENSE.txt for more info.
*		
*	Filename	-	ParameterSystem.cpp
*
*	Comments	-	See ParameterSystem.h
*
**************************************************************************************/
#include "../Include/ParameterSystem.h"

namespace Pulse
{
	ParameterSystem::ParameterSystem( void )
	{
	}

	ParameterSystem::~ParameterSystem(	void )
	{
		ClearData();
	}

	IParameter * ParameterSystem::GetParameter( const CHAR *pName )
	{
		RenderParameterList::Iterator iter = m_renderParams.Find( pName );
		
		if ( iter != m_renderParams.IteratorEnd() )
			return iter->second;

		return PSX_NULL;
	}

	ParameterScalar * ParameterSystem::GetScalar( const CHAR *pName )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert(pParam->GetType() == EParameter::SCALAR, "Parameter should be a SCALAR type." );
			return (ParameterScalar*)pParam;
		}

		pParam = new ParameterScalar;
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterScalar*)pParam;
	}

	ParameterVector3 * ParameterSystem::GetVector3( const CHAR *pName )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert(pParam->GetType() == EParameter::VECTOR3, "Parameter should be a vector type." );
			return (ParameterVector3*)pParam;
		}

		pParam = new ParameterVector3;
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterVector3*)pParam;
	}

	ParameterVector4 * ParameterSystem::GetVector4( const CHAR *pName )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert(pParam->GetType() == EParameter::VECTOR4, "Parameter should be a vector type." );
			return (ParameterVector4*)pParam;
		}

		pParam = new ParameterVector4;
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterVector4*)pParam;
	}

	ParameterMatrix * ParameterSystem::GetMatrix( const CHAR *pName )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert(pParam->GetType() == EParameter::MATRIX, "Parameter should be a matrix type." );
			return (ParameterMatrix*)pParam;
		}

		pParam = new ParameterMatrix;
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterMatrix*)pParam;
	}

	ParameterMatrixArray * ParameterSystem::GetMatrixArray( const CHAR *pName, UINT numMatrices )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert(pParam->GetType() == EParameter::MATRIX_ARRAY, "Parameter should be a matrix array type." );
			if ( ((ParameterMatrixArray*)pParam)->GetNumMatrices() == numMatrices )
				return (ParameterMatrixArray*)pParam;
		}

		pParam = new ParameterMatrixArray(numMatrices);
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterMatrixArray*)pParam;
	}

	ParameterShaderResource * ParameterSystem::GetShaderResource( const CHAR *pName, EShaderResource::Type type )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert(pParam->GetType() == EParameter::SHADER_RESOURCE && ((ParameterShaderResource*)pParam)->GetShaderResourceType() == type, "Parameter should be a shader resource type." );
			PSX_Assert( ((ParameterShaderResource*)pParam)->GetShaderResourceType() != EShaderResource::UNKNOWN, "Shader resource type shouldn't be unknown." );
			return (ParameterShaderResource*)pParam;
		}

		PSX_Assert( type != EShaderResource::UNKNOWN, "Shader resource type shouldn't be unknown." );

		pParam = new ParameterShaderResource( type );
		pParam->SetName( pName );
		//((ParameterShaderResource*)pParam)->GetResource()->SetShaderResourceType( type );
		m_renderParams[pName] = pParam;
		return (ParameterShaderResource*)pParam;
	}

	ParameterConstantBuffer * ParameterSystem::GetConstantBuffer( const CHAR *pName )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert( pParam->GetType() == EParameter::CONSTANT_BUFFER, "Parameter should be a constant buffer type." );
			return (ParameterConstantBuffer*)pParam;
		}

		pParam = new ParameterConstantBuffer;
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterConstantBuffer*)pParam;
	}

	ParameterSamplerState * ParameterSystem::GetSamplerState( const CHAR *pName )
	{
		IParameter *pParam = GetParameter( pName );

		if ( pParam )
		{
			PSX_Assert( pParam->GetType() == EParameter::SAMPLER, "Parameter should be a Sampler state type." );
			return (ParameterSamplerState*)pParam;
		}

		pParam = new ParameterSamplerState;
		pParam->SetName( pName );
		m_renderParams[pName] = pParam;
		return (ParameterSamplerState*)pParam;
	}

	void ParameterSystem::ClearData( void )
	{
		RenderParameterList::Iterator iter = m_renderParams.IteratorBegin();
		RenderParameterList::Iterator iterEnd = m_renderParams.IteratorEnd();

		while ( iter != iterEnd )
		{
			PSX_SafeDelete( iter->second );
			++iter;
		}

		m_renderParams.Clear();
	}

}